/******************************
* Copyright (c) 2003,2019 Kevin Lano
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0
*
* SPDX-License-Identifier: EPL-2.0
* *****************************/
/* Package: Code Generation */ 

import java.util.Vector; 

public class CGRule
{ String lhs;
  String rhs;
  Vector variables; // The _i in lhs -- no new _i should be in rhs
  Vector metafeatures; // The _i`f in rhs
  String lhsop = "";
  Vector conditions;

  public CGRule(Expression lexp, Expression rexp, Vector whens)
  { Vector lvars = lexp.metavariables();
    Vector rvars = rexp.metavariables();
    
    if (lvars.containsAll(rvars)) {}
    else
    { System.err.println("!! Error: some extra metavariables on RHS of " + lexp + " |--> " + rexp); }
    lhs = lexp + "";
    rhs = rexp + "";
    variables = lvars;
    conditions = whens;
    metafeatures = metafeatures(rhs); 
  }

  public CGRule(Expression lexp, String rgt, Vector whens)
  { variables = lexp.metavariables();
    lhs = lexp + "";
    rhs = rgt;

    Vector rvariables = metavariables(rgt); 
    if (variables.containsAll(rvariables)) { } 
    else 
    { System.err.println("!! Error: some extra metavariables on RHS of " + this); }

    conditions = whens;
    metafeatures = metafeatures(rhs); 
  }

  public CGRule(Expression lexp, String rgt)
  { variables = lexp.metavariables();
    lhs = lexp + "";
    rhs = rgt;

    Vector rvariables = metavariables(rgt); 
    if (variables.containsAll(rvariables)) { } 
    else 
    { System.err.println("!! Error: some extra metavariables on RHS of " + this); }

    conditions = new Vector();
    metafeatures = metafeatures(rhs); 
  }

  public CGRule(Statement lexp, String rgt, Vector whens)
  { variables = lexp.metavariables();
    lhs = lexp + "";
    rhs = rgt;

    Vector rvariables = metavariables(rgt); 
    if (variables.containsAll(rvariables)) { } 
    else 
    { System.err.println("!! Error: some extra metavariables on RHS of " + this); }

    conditions = whens;
    metafeatures = metafeatures(rhs); 
  }

  public CGRule(Statement lexp, String rgt)
  { variables = lexp.metavariables();
    lhs = lexp + "";
    rhs = rgt;

    Vector rvariables = metavariables(rgt); 
    if (variables.containsAll(rvariables)) { } 
    else 
    { System.err.println("!! Error: some extra metavariables on RHS of " + this); }

    conditions = new Vector();
    metafeatures = metafeatures(rhs); 
  }

  public CGRule(Type lexp, String rgt, Vector whens)
  { variables = lexp.metavariables();
    lhs = lexp + "";
    rhs = rgt;

    Vector rvariables = metavariables(rgt); 
    if (variables.containsAll(rvariables)) { } 
    else 
    { System.err.println("!! Error: some extra metavariables on RHS of " + this); }

    conditions = whens;
    metafeatures = metafeatures(rhs); 
  }


  public CGRule(Type lexp, String rgt)
  { variables = lexp.metavariables();
    lhs = lexp + "";
    rhs = rgt;

    Vector rvariables = metavariables(rgt); 
    if (variables.containsAll(rvariables)) { } 
    else 
    { System.err.println("!! Error: some extra metavariables on RHS of " + this); }

    conditions = new Vector();
    metafeatures = metafeatures(rhs); 
  }

  public CGRule(String ls, String rs, Vector vs, Vector whens)
  { lhs = ls;
    rhs = rs;
    variables = vs;

    Vector rvariables = metavariables(rs); 
    if (variables.containsAll(rvariables)) { } 
    else 
    { System.err.println("!! Error: some extra metavariables on RHS of " + this); }

    conditions = whens;
    metafeatures = metafeatures(rhs); 
  }

  public boolean hasVariables()
  { return variables != null && variables.size() > 0; } 

  public static Vector metavariables(String str) 
  { Vector res = new Vector(); 
    for (int i = 1; i < 10; i++) 
    { String var = "_" + i; 
      if (str.indexOf(var) > -1) 
      { res.add(var); } 
    } 
    return res; 
  } 

  public static Vector metafeatures(String str) 
  { Vector res = new Vector(); 
    for (int i = 1; i < 10; i++) 
    { String var = "_" + i + "`"; 
      if (str.indexOf(var) > -1) 
      { int j = str.indexOf(var); 
        String f = var; 
        boolean found = false; 
        for (int k = j+3; k < str.length() && !found; k++) 
        { if (Character.isLetter(str.charAt(k)))
          { f = f + str.charAt(k); } 
          else 
          { res.add(f);
            found = true; 
          } 
        } 
      } 
    } 
    return res; 
  } // look for identifier starting from the `

  public String toString() 
  { String res = lhs + " |-->" + rhs; 
    if (conditions != null && conditions.size() > 0) 
    { res = res + "<when> "; 
      for (int i = 0; i < conditions.size(); i++) 
      { CGCondition cnd = (CGCondition) conditions.get(i); 
        res = res + cnd;
        if (i < conditions.size() - 1) 
        { res = res + ", "; }
      } 
    }
    return res;  
  } 
 

  public void addCondition(CGCondition cond)
  { conditions.add(cond); }

  public boolean hasCondition(String prop)
  { for (int x = 0; x < conditions.size(); x++)
    { CGCondition cond = (CGCondition) conditions.get(x);
      if (prop.equals(cond.stereotype) && cond.positive)
      { return true; }
    }
    return false;
  }

  public boolean hasCondition(String prop, String var)
  { for (int x = 0; x < conditions.size(); x++)
    { CGCondition cond = (CGCondition) conditions.get(x);
      if (prop.equals(cond.stereotype) && var.equals(cond.variable) && cond.positive)
      { return true; }
    }
    return false;
  }

  public boolean satisfiesConditions(Vector args)
  { return CGCondition.conditionsSatisfied(conditions,args); } 

  public String applyRule(Vector args)
  { // substitute variables[i] by args[i] in rhs
    String res = rhs + "";
    for (int x = 0; x < args.size() && x < variables.size(); x++)
    { String var = (String) variables.get(x);
      String arg = (String) args.get(x);
      String arg1 = correctNewlines(arg); 
      System.out.println(">--> Replacing " + var + " by " + arg1); 
      res = res.replaceAll(var,arg1);
    }
    return res;
  }

  public String applyRule(Vector args, Vector eargs, CGSpec cgs)
  { // substitute metafeatures[j] by the value of eargs[j] metafeature
    // substitute variables[i] by args[i] in rhs
    
    String res = rhs + "";
    for (int j = 0; j < metafeatures.size(); j++) 
    { String mf = (String) metafeatures.get(j); 
      String mfvar = mf.substring(0,2); 
      String mffeat = mf.substring(3,mf.length());
      int k = Integer.parseInt(mfvar.charAt(1) + "");  
      if (k >= 1 && k <= eargs.size())
      { Object obj = eargs.get(k-1);
        if ("elementType".equals(mffeat) && obj instanceof Expression)
        { Expression e = (Expression) obj; 
          Type t = e.getElementType(); 
          if (t != null) 
          { String repl = t.cg(cgs); 
            String repl1 = correctNewlines(repl); 
            System.out.println(">--> Replacing " + mf + " by " + repl1); 
            res = res.replaceAll(mf,repl1);
          } 
        }
        else if ("type".equals(mffeat) && obj instanceof Expression)
        { Expression e = (Expression) obj; 
          Type t = e.getType(); 
          if (t != null) 
          { String repl = t.cg(cgs); 
            String repl1 = correctNewlines(repl); 
            System.out.println(">--> Replacing " + mf + " by " + repl1); 
            res = res.replaceAll(mf,repl1);
          } 
        } 
        else if ("name".equals(mffeat) && obj instanceof ModelElement)
        { ModelElement e = (ModelElement) obj; 
          String repl = e.getName(); 
          if (repl != null) 
          { String repl1 = correctNewlines(repl); 
            System.out.println(">--> Replacing " + mf + " by " + repl1); 
            res = res.replaceAll(mf,repl1);
          } 
        } 
      } 
    }
 

    for (int x = 0; x < args.size() && x < variables.size(); x++)
    { String var = (String) variables.get(x);
      String arg = (String) args.get(x);
      String arg1 = correctNewlines(arg); 
      System.out.println(">--> Replacing " + var + " by " + arg1); 
      res = res.replaceAll(var,arg1);
    }
    return res;
  }

  private String correctNewlines(String str) 
  { String res = ""; 
    if (str.length() == 0) 
    { return res; } 

    for (int i = 0; i < str.length() - 1; i++) 
    { char c1 = str.charAt(i); 
      char c2 = str.charAt(i+1); 
      if (c1 == '\\' && c2 == 'n')
      { res = res + '\n'; 
        i++;
        if (i == str.length() - 1)
        { return res; }  
      } 
      else 
      { res = res + c1; } 
    } 
    return res + str.charAt(str.length()-1); 
  } 

  public static void main(String[] args) 
  { System.out.println(metafeatures("for (_1`elementType _2 : _1) do { _3 }")); } 
}
